from cytoolz import first
from cytoolz import second
from merlin import cfg
from merlin import chips
from merlin import dates
from merlin import functions as f
from merlin import specs
from merlin import rods
from functools import reduce
import numpy as np
import test
 
 
def test_from_chips():
    chips = list()
    chips.append({'data': np.int_([[11, 12, 13],
                                   [14, 15, 16],
                                   [17, 18, 19]])})
    chips.append({'data': np.int_([[21, 22, 23],
                                   [24, 25, 26],
                                   [27, 28, 29]])})
    chips.append({'data': np.int_([[31, 32, 33],
                                   [34, 35, 36],
                                   [37, 38, 39]])})
    pillar = rods.from_chips(chips)
    assert pillar.shape[0] == chips[0]['data'].shape[0]
    assert pillar.shape[1] == chips[0]['data'].shape[1]
    assert pillar.shape[2] == len(chips)

    # going to flatten both the chips arrays and the pillar array
    # then perform black magic and verify that the values wound up
    # where they belonged in the array positions.
    # using old style imperative code as double insurance
    # happiness is not doing this and relying on functional principles only
    # because direct manipulation of memory locations is error prone and
    # very difficult to think about.
    # We're mapping array locations between two things that look like this:
    # [11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27 ...]
    # [11, 21, 31, 12, 22, 32, 13, 23, 33, 14, 24, 34, 15, 25, 35, 16, 26 ...]
    # This alone should serve as all the evidence needed to prove that
    # imperative programming is bad for everyone involved.  I am very sorry.
    fchips = list(f.flatten([c['data'].flatten() for c in chips]))
    jump = reduce(lambda accum, v: accum + v, pillar.shape) # 9
    modulus = pillar.shape[0] # 3
    for i, val in enumerate(pillar.flatten()):
        factor = i % modulus # 0 1 2
        group = i // modulus # 0 1 2
        assert val == fchips[(factor * jump) + group]


def test_locate():
    # test data.  sum(locs) + 2 should equal sum(rods) per inner array
    # value of sum of every inner array element should be unique.
    _locs = np.int_([[[0, 0], [0, 1], [0, 2]],
                     [[1, 3], [1, 4], [1, 5]],
                     [[2, 6], [2, 7], [2, 8]]])

    _rods = np.int_([[[0, 0, 2], [0, 0, 3], [0, 0, 4]],
                     [[1, 0, 5], [1, 0, 6], [1, 0, 7]],
                     [[2, 0, 8], [2, 0, 9], [2, 0, 10]]])

    # sanity check
    locs_total = _locs.reshape(9, 2).sum(axis=1) + 2
    rods_total = _rods.reshape(9, 3).sum(axis=1)
    assert np.array_equal(locs_total, rods_total)

    # make sure we located all the rods correctly.
    flatlocs = _locs.reshape(9, 2)
    tseries = rods.locate(locations=_locs, rods=_rods)
    assert all([tseries[tuple(loc)].sum() == loc.sum() + 2 for loc in flatlocs])


@test.vcr.use_cassette(test.cassette)
def test_create():
 
    c = cfg.get('chipmunk-ard', env=test.ard_env)
 
    x, y = c.get('snap_fn')(x=test.x, y=test.y).get('chip').get('proj-pt')

    ubids = cfg.ubids.get('chipmunk-ard').get('reds')

    registry = c.get('registry_fn')()
    
    refspec = specs.refspec(specs.mapped(specs=registry, ubids={'reds': ubids}))
    # print("REFSPEC:{}".format(refspec))
 
    chipseq = c.get('chips_fn')(x=x,
                                y=y,
                                acquired=test.ard_acquired,
                                ubids=ubids)

    dateseq = dates.mapped(chipmap=dict(reds=chipseq)).get('reds')

    grid = {x['name']: x for x in c.get('grid_fn')()}.get('chip')

    decode_fn = c.get('decode_fn')
 
    locations = chips.locations(x=x,
                                y=y,
                                cw=first(refspec.get('data_shape')),
                                ch=second(refspec.get('data_shape')),
                                rx=grid.get('rx'),
                                ry=grid.get('ry'),
                                sx=grid.get('sx'),
                                sy=grid.get('sy'))

    _rods = rods.create(x=x,
                        y=y,
                        chipseq=chipseq,
                        dateseq=dateseq,
                        locations=locations,
                        spec_index=specs.index(registry),
                        decode_fn=decode_fn)

    assert len(_rods) == 10000
    assert type(_rods) is dict



from cytoolz import first
import cytoolz
import requests


def only(ubids, specs):
    """Filter specs on ubids.

    Args:
        ubids (seq): [ubid1, ubid3]
        specs (seq): [{spec1}, {spec2}, {spec3}, ...})

    Returns:
        tuple: [{spec1}, {spec3}]
    """

    return tuple(cytoolz.filter(lambda x: x['ubid'] in ubids, specs))


def mapped(ubids, specs):
    """Organizes specs by key.

    Args:
        ubids (dict): {'reds':   ['ubid1', 'ubid2'], 
                       'greens': ['ubid3', 'ubid4']}
        specs (seq):  [{spec1}, {spec2}, {spec3}, {spec4}]

    Returns:
        dict

    Example:
        >>> {'reds':   [spec1, spec2], 
             'greens': [spec3, spec4]}
    """
    
    return {k: only(v, specs) for k, v in ubids.items()}


def exist(ubids, specs):
    """Checks that all ubids are in the specs

    Args:
        ubids (seq): [ubid1, ubid2, ubid3, ...]
        specs (seq): [{spec1}, {spec2}, {spec3}]

    Returns:
        bool: True or False
    """

    return set(ubids).issubset(set(map(lambda x: x['ubid'], specs)))
    

def index(specs):
    """Organizes specs by ubid

    Args:
        specs (sequence): a sequence of chip specs

    Returns:
        dict: specs keyed by ubid
    """

    return {s['ubid']: s for s in specs}


def ubids(specs):
    """Extract ubids from a sequence of specs

    Args:
        specs (sequence): a sequence of spec dicts

    Returns:
        tuple: a sequence of ubids
    """

    return tuple(s['ubid'] for s in specs if 'ubid' in s)


def refspec(specmap):
    """Returns the first chip spec from the first key to use as a reference.

    Args:
        specmap: {key: [specs]}

    Returns:
        dict: spec
    """

    return first(specmap[first(specmap)])



from cytoolz import merge
from cytoolz import partial
from merlin import chips
from merlin import dates
from merlin import functions
from merlin import rods
from merlin import specs


def pyccd(x, y, locations, dates_fn, specmap, chipmap, decode_fn):
    """Builds inputs for the pyccd algorithm.

    Args:
        x: x projection coordinate of chip
        y: y projection coordinate of chip
        locations: chip shaped 2d array of projection coordinates
        dates_fn (fn): returns dates that should be included in time series
        specmap (dict): mapping of keys to specs
        chipmap (dict): mapping of keys to chips
        decode_fn (fn): Function that accepts a chip and returns a decoded chip
                        (See chips.decode)

    Returns:
        A tuple of tuples.

    The pyccd format key is ```(chip_x, chip_y, x, y)``` with a
    dictionary of sorted numpy arrays representing each spectra plus an
    additional sorted dates array.

    >>> pyccd_format(*args)
        (((chip_x, chip_y, x1, y1), {"dates":  [],  "reds": [],
                                     "greens": [],  "blues": [],
                                     "nirs1":  [],  "swir1s": [],
                                     "swir2s": [],  "thermals": [],
                                     "qas":    []}),
         ((chip_x, chip_y, x1, y2), {"dates":  [],  "reds": [],
                                     "greens": [],  "blues": [],
                                     "nirs1":  [],  "swir1s": [],
                                     "swir2s": [],  "thermals": [],
                                     "qas":    []}))
        ...
    """
    
    _index   = specs.index(list(functions.flatten(specmap.values())))
    _dates   = dates_fn(datemap=dates.mapped(chipmap))
    _creator = partial(rods.create,
                       x=x,
                       y=y,
                       dateseq=_dates,
                       locations=locations,
                       spec_index=_index,
                       decode_fn=decode_fn)
    _flipped = partial(functions.flip_keys, {k: _creator(chipseq=v) for k, v in chipmap.items()})
    
    _rods = functions.insert_into_every(key='dates',
                                        value=list(map(dates.to_ordinal, dates.rsort(_dates))),
                                        dods=_flipped())
                                 
    return tuple((k, v) for k, v in _rods.items())


def aux(x, y, locations, dates_fn, specmap, chipmap, decode_fn):
    """Builds aux formatted timeseries.

    Args:
        x: x projection coordinate of chip
        y: y projection coordinate of chip
        locations: chip shaped 2d array of projection coordinates
        dates_fn (fn): returns dates that should be included in time series
        specmap (dict): mapping of keys to specs
        chipmap (dict): mapping of keys to chips
        decode_fn (fn): Function that accepts a chip and returns a decoded chip
                        (See chips.decode)

    Returns:
        A tuple of tuples.

    The aux format key is ```(chip_x, chip_y, x, y)``` with a
    dictionary of sorted numpy arrays representing each spectra plus an
    additional sorted dates array.

    >>> aux(*args)
        (((chip_x, chip_y, x1, y1), {"dates":   [], "nlcd": [],
                                     "nlcdtrn": [], "trends": [],
                                     "mpw":     [], "dem": [],
                                     "posidex": [], "aspect": [],
                                     "slope":   []}),
         ((chip_x, chip_y, x1, y2), {"dates":   [], "nlcd": [],
                                     "nlcdtrn": [], "trends": [],
                                     "mpw":     [], "dem": [],
                                     "posidex": [], "aspect": [],
                                     "slope":   []}))
        ...
    """
    
    _index   = specs.index(list(functions.flatten(specmap.values())))
    _dates   = dates_fn(datemap=dates.mapped(chipmap))
    _creator = partial(rods.create,
                       x=x,
                       y=y,
                       dateseq=_dates,
                       locations=locations,
                       spec_index=_index,
                       decode_fn=decode_fn)
    _flipped = partial(functions.flip_keys, {k: _creator(chipseq=v) for k, v in chipmap.items()})
    _rods = functions.insert_into_every(key='dates',
                                        value=[dates.minmax(_dates)],
                                        dods=_flipped())
    return tuple((k, v) for k, v in _rods.items())





""" Functions for working with files in python """
from merlin import files


def test_read():
    assert 1 > 0


def test_readb():
    assert 1 > 0


def test_readlines():
    assert 1 > 0


def test_readlinesb():
    assert 1 > 0


def test_write():
    assert 1 > 0


def test_writeb():
    assert 1 > 0


def test_append():
    assert 1 > 0


def test_appendb():
    assert 1 > 0


def test_delete():
    assert 1 > 0


def test_exists():
    assert 1 > 0



"""Functions for working with files in python"""

import logging
import os

logger = logging.getLogger(__name__)

def read(path):
    """Read a text file.

    Args:
        path (str): Full path to file

    Returns:
        str: File text
    """

    with open(path, 'r') as handle:
        return handle.read()


def readb(path):
    """Read a binary file.

    Args:
        path (str): Full path to file

    Returns:
        File bytes
    """

    with open(path, 'rb') as handle:
        return handle.read()


def readlines(path):
    """Read lines from a text file.

    Args:
        path (str): Full path to file

    Returns:
        list: File text
    """

    with open(path, 'r') as handle:
        return handle.readlines()


def readlinesb(path):
    """Read lines from a binary file.

    Args:
        path (str): Full path to file

    Returns:
        list: File bytes
    """

    with open(path, 'rb') as handle:
        return handle.readlines()


def write(path, data):
    """Write data to a text file.

    Args:
        path (str): Full path to file
        data (str): File text

    Returns:
        int: Number of characters written
    """

    with open(path, 'w') as handle:
        return handle.write(data)


def writeb(path, data):
    """Write data to a binary file.

    Args:
        path (str): Full path to file
        data (str): File bytes

    Returns:
        int: Number of bytes written
    """

    with open(path, 'wb') as handle:
        return handle.write(data)


def append(path, data):
    """Append data to a text file.

    Args:
        path (str): Full path to file
        data (str): File text

    Returns:
        int: Number of characters appended
    """

    with open(path, 'a') as handle:
        return handle.write(data)


def appendb(path, data):
    """Write data to a binary file.

    Args:
        path (str): Full path to file
        data (str): File bytes

    Returns:
        int: Number of bytes appended
    """

    with open(path, 'ab') as handle:
        return handle.write(data)


def delete(path):
    """Delete a file.

    Args:
        path (str): Full path to file

    Returns:
        bool: True if the file was deleted, False if not (with message logged)
    """

    try:
        os.remove(path)
    except FileNotFoundError:
        pass
    except Exception as e:
        logger.error("Exception deleting file:{}".format(path))
        logger.error(e)
        return False
    return True


def exists(path):
    """Determine if a file exists.

    Args:
        path (str): Full path to file

    Returns:
        bool: True if the file exists, False if not
    """

    return os.path.exists(path) and os.path.isfile(path)


def mkdirs(filename):
    """Ensures the set of directories exist for the supplied filename.

    Args:
        filename (str): Full path to where the file should exist

    Returns:
       Full file path if the directories were created or None
    """

    try:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        return filename
    except:
        return None



from cytoolz import concat
from cytoolz import first
from cytoolz import identity
from merlin import cfg
from merlin import chipmunk
import test

@test.vcr.use_cassette(test.cassette)
def test_chips():
    red_chips = chipmunk.chips(x=test.x,
                               y=test.y,
                               acquired=test.ard_acquired,
                               ubids=cfg.ubids.get('chipmunk-ard').get('reds'),
                               url=test.ard_env.get('CHIPMUNK_URL'))
    assert len(red_chips) > 0
    assert type(red_chips) is tuple
    

@test.vcr.use_cassette(test.cassette)
def test_registry():

    registry = chipmunk.registry(url=test.ard_env.get('CHIPMUNK_URL'))
    assert len(registry) > 0
    assert type(registry) is tuple or list

    entry = first(registry)
    assert type(entry) is dict

    keys = ['ubid', 'info', 'tags', 'data_type', 'data_fill', 'data_shape']
    assert all([key in entry for entry in registry for key in keys])
    

@test.vcr.use_cassette(test.cassette)
def test_grid():
    grid = chipmunk.grid(url=test.ard_env.get('CHIPMUNK_URL'))
    assert len(grid) == 2
    assert type(grid) is list or tuple
    assert set(['chip', 'tile']) == set([g.get('name') for g in grid])

    keys = ['proj', 'rx', 'ry', 'sx', 'sy', 'tx', 'ty']
    assert all([key in g for g in grid for key in keys])
    

@test.vcr.use_cassette(test.cassette)
def test_snap():
    snapped = chipmunk.snap(x=test.x,
                            y=test.y,
                            url=test.ard_env.get('CHIPMUNK_URL'))
    
    assert type(snapped) is dict
    assert len(snapped) is 2
    
    keys = ['chip', 'tile']
    assert all([key in snapped for key in keys])

    subkeys = ['grid-pt', 'proj-pt']
    assert all([subkey in v for k, v in snapped.items() for subkey in subkeys])


@test.vcr.use_cassette(test.cassette)
def test_near():
    near = chipmunk.near(x=test.x,
                         y=test.y,
                         url=test.ard_env.get('CHIPMUNK_URL'))
    
    assert type(near) is dict

    assert len(near) is 2
    
    keys = ['chip', 'tile']

    assert all([key in near for key in keys])

    assert all([len(near[key]) == 9 for key in keys])

    assert all([type(near[key]) is list or tuple for key in keys])
    
    assert all(map(lambda key: key in near, keys))

    assert all(map(lambda sub: type(sub) is list or tuple, map(lambda key: near[key], keys)))

    assert all(list(map(lambda entry: 'grid-pt' in entry and 'proj-pt' in entry,
                        concat(map(lambda key: near[key], keys)))))
    
 



"""Chipmunk.py is the interface module to Chipmunk for Merlin.

Any functions chipmunk exposes are represented here as defined.

New abstractions or higher level abstractions should be created
in modules that import chipmunk.py.  For maximum flexibility do not
import chipmunk.py directly... inject it via a kernel or DI construct.  

Multi-version support: 
To support multiple versions of Chipmunk create new modules that correspond to
the appropriate version number.
"""

from cytoolz import get
from cytoolz import reduce
from functools import partial
from operator import add

import logging
import requests

logger = logging.getLogger(__name__)

def chips(x, y, acquired, ubids, url, resource='/chips'):
    """Returns chips from a Chipmunk url given x, y, date range and ubid sequence

    Args:
        x (int): projection coordinate x
        y (int): projection coordinate y
        acquired (str): ISO8601 daterange '2012-01-01/2014-01-03'
        ubids (sequence): sequence of ubids
        url (str): protocol://host:port/path
        resource (str): /chips/resource/path (default: /chips)

    Returns:
        tuple: chips

    Example:
        >>> chipmunk.chips(url='http://host:port/path',
                           x=123456,
                           y=789456,
                           acquired='2012-01-01/2014-01-03',
                           ubids=['LE07_SRB1', 'LT05_SRB1'])
        (LE07_SRB1_DATE1, LT05_SRB1_DATE2, LE07_SRB1_DATE2, ...)
    """

    url = '{}{}'.format(url, resource)

    params = [{'x': x, 'y': y, 'acquired': acquired, 'ubid': u } for u in ubids]
    
    def request(url, params):
        r = requests.get(url=url, params=params)

        if not r.ok:
            logger.error("{} at {} for {}".format(r.reason, url, params))
            logger.debug(r.text)
            return None
        else:
            return r.json()
    
    responses = [request(url=url, params=p) for p in params]
    return tuple(reduce(add, filter(lambda x: type(x) in [list, tuple], responses), []))
              

def registry(url, resource='/registry'):
    """Retrieve the chip spec registry

    Args:
        url (str): protocol://host:port/path
        resource (str): /registry/resource/path (default: /registry)

    Returns:
        list

    Example:
        >>> chipmunk.registry(url='http://host:port/path')
         [{'data_fill': '-9999',
           'data_mask': {},
           'data_range': [],
           'data_scale': None,
           'data_shape': [100, 100],
           'data_type': 'INT16',
           'data_units': None,
           'info': 'band 5 top-of-atmosphere reflectance',
           'tags': ['swir1', 'b5', 'tab5', 'lt05', 'lt05_tab5', 'ta'],
           'ubid': 'LT05_TAB5'},
          {'data_fill': '-9999',
           'data_mask': {},
           'data_range': [],
           'data_scale': None,
           'data_shape': [100, 100],
           'data_type': 'INT16',
           'data_units': None,
           'info': 'band 7 top-of-atmosphere reflectance',
           'tags': ['lt05_tab7', 'b7', 'lt05', 'swir2', 'tab7', 'ta'],
           'ubid': 'LT05_TAB7'}, ...]
    """
    
    return requests.get(url="{}{}".format(url, resource)).json()


def grid(url, resource='/grid'):
    """Return grid definitions.
  
    Args:
        url (str): protocol://host:port/path
        resource (str): /the/grid/resource (default: /grid)
 
    Returns:
        dict

    Example:
        >>> chipmunk.grid(url='http://host:port/path)
        [{"name":"tile",
          "proj":null,
          "rx":1.0,
          "ry":-1.0,
          "sx":150000.0,
          "sy":150000.0,
          "tx":2565585.0,
          "ty":3314805.0},
         {"name":"chip",
          "proj":null,
          "rx":1.0,
          "ry":-1.0,
          "sx":3000.0,
          "sy":3000.0,
          "tx":2565585.0,
          "ty":3314805.0}]
    """
    
    url = '{}{}'.format(url, resource)
    return requests.get(url=url).json()


def snap(x, y, url, resource='/grid/snap'):
    """Determine the chip and tile coordinates for a point.
  
    Args:
        x (int): projection coordinate x
        y (int): projection coordinate y
        url (str): protocol://host:port/path
        resource (str): /grid/snap/resource (default: /grid/snap)
 
    Returns:
        dict

    Example:
        >>> chipmunk.snap(x=0, y=0, url='http://host:port/path')
         {'chip': {'grid-pt': [855.0, 1104.0], 'proj-pt': [-585.0, 2805.0]},
          'tile': {'grid-pt': [17.0, 22.0], 'proj-pt': [-15585.0, 14805.0]}}
    """
    
    url = '{}{}'.format(url, resource)
    return requests.get(url=url, params={'x': x, 'y': y}).json()


def near(x, y, url, resource='/grid/near'):
    """Determines chips and tiles that lie a point

    Args:
        x (int): projection coordinate x
        y (int): projection coordinate y
        url (str): protocol://host:port/path
        resource (str): /grid/near/resource (default: /grid/near)
    
    Returns:
        dict

    Example:
        >>> chipmunk.near(x=0, y=0, url='http://host:port/path')
        {'chip': [{'grid-pt': [854.0, 1105.0], 'proj-pt': [-3585.0, -195.0]},
                  {'grid-pt': [854.0, 1104.0], 'proj-pt': [-3585.0, 2805.0]},
                  {'grid-pt': [854.0, 1103.0], 'proj-pt': [-3585.0, 5805.0]},
                  {'grid-pt': [855.0, 1105.0], 'proj-pt': [-585.0, -195.0]},
                  {'grid-pt': [855.0, 1104.0], 'proj-pt': [-585.0, 2805.0]},
                  {'grid-pt': [855.0, 1103.0], 'proj-pt': [-585.0, 5805.0]},
                  {'grid-pt': [856.0, 1105.0], 'proj-pt': [2415.0, -195.0]},
                  {'grid-pt': [856.0, 1104.0], 'proj-pt': [2415.0, 2805.0]},
                  {'grid-pt': [856.0, 1103.0], 'proj-pt': [2415.0, 5805.0]}],
         'tile': [{'grid-pt': [16.0, 23.0], 'proj-pt': [-165585.0, -135195.0]},
                  {'grid-pt': [16.0, 22.0], 'proj-pt': [-165585.0, 14805.0]},
                  {'grid-pt': [16.0, 21.0], 'proj-pt': [-165585.0, 164805.0]},
                  {'grid-pt': [17.0, 23.0], 'proj-pt': [-15585.0, -135195.0]},
                  {'grid-pt': [17.0, 22.0], 'proj-pt': [-15585.0, 14805.0]},
                  {'grid-pt': [17.0, 21.0], 'proj-pt': [-15585.0, 164805.0]},
                  {'grid-pt': [18.0, 23.0], 'proj-pt': [134415.0, -135195.0]},
                  {'grid-pt': [18.0, 22.0], 'proj-pt': [134415.0, 14805.0]},
                  {'grid-pt': [18.0, 21.0], 'proj-pt': [134415.0, 164805.0]}]}
    """
    
    url = '{}{}'.format(url, resource)
    return requests.get(url=url, params={'x': x, 'y': y}).json()



from base64 import b64decode
from base64 import b64encode
from cytoolz import concat
from cytoolz import drop
from cytoolz import first
from cytoolz import filter
from merlin import cfg
from merlin import chips
from merlin import specs
from functools import partial
from functools import reduce
from itertools import product
from numpy.random import randint
import numpy as np
import test

   
def test_locations():
    params = {'x': 0,
              'y': 0,
              'cw': 2,
              'ch': 2,
              'rx': 1,
              'ry': -1,
              'sx': 60,
              'sy': 60}

    locs = np.array([[[0, 0], [30, 0]], [[0, -30], [30, -30]]])
    assert np.array_equal(locs, chips.locations(**params))
    

def test_dates():
    inputs = list()
    inputs.append({'acquired': '2015-04-01'})
    inputs.append({'acquired': '2017-04-01'})
    inputs.append({'acquired': '2017-01-01'})
    inputs.append({'acquired': '2016-04-01'})
    assert set(chips.dates(inputs)) == set(map(lambda d: d['acquired'], inputs))


def test_trim():
    inputs = list()
    inputs.append({'include': True, 'acquired': '2015-04-01'})
    inputs.append({'include': True, 'acquired': '2017-04-01'})
    inputs.append({'include': False, 'acquired': '2017-01-01'})
    inputs.append({'include': True, 'acquired': '2016-04-01'})
    included = chips.dates(filter(lambda d: d['include'] is True, inputs))
    trimmed = chips.trim(dates=included, chips=inputs)
    assert len(list(trimmed)) == len(included)
    assert set(included) == set(map(lambda x: x['acquired'], trimmed))


def test_decode():
    input_data  = b'[1, 2, 3, 4]'
    input_chip  = {'data': b64encode(input_data)}
    output_chip = chips.decode(input_chip)
    assert input_data == output_chip['data']

    
def test_chip_to_numpy():
    data_type = 'int16'
    data_shape = [2, 2]

    input_data = np.array([1, 2, 3, 4], dtype=data_type).reshape(data_shape)
    input_chip = {'data': input_data}
    
    output_chip = chips.chip_to_numpy(chip=input_chip,
                                      chip_spec={'data_shape': data_shape,
                                                 'data_type': data_type})
    output_data = output_chip.get('data')
    
    assert np.array_equal(input_data, output_data)
   

def test_to_numpy():
    """ Builds combos of shapes and numpy data types and tests
        chipmunk.to_numpy() with all of them """

    def _ubid(dtype, shape):
        return dtype + str(shape)

    def _chip(dtype, shape, ubid):
        limits = np.iinfo(dtype)
        length = reduce(lambda accum, v: accum * v, shape)
        matrix = randint(limits.min, limits.max, length, dtype).reshape(shape)
        return {'ubid': ubid, 'data': matrix}

    def _spec(dtype, shape, ubid):
        return {'ubid': ubid, 'data_shape': shape, 'data_type': dtype.upper()}

    def _check(npchip, spec_index):
        spec = spec_index[npchip['ubid']]
        assert npchip['data'].dtype.name == spec['data_type'].lower()
        assert npchip['data'].shape == spec['data_shape']
        return True

    # Test combos of dtypes/shapes to ensure data shape and type are unaltered
    combos = tuple(product(('uint8', 'uint16', 'int8', 'int16'),
                           ((3, 3), (1, 1), (100, 100))))

    # generate the chip_specs and chips
    _chips = [_chip(*c, _ubid(*c)) for c in combos]
    _specs = [_spec(*c, _ubid(*c)) for c in combos]
    _spec_index = specs.index(_specs)

    # run assertions
    checker = partial(_check, spec_index=_spec_index)
    all(map(checker, chips.to_numpy(spec_index=_spec_index, chips=_chips)))


def test_identity():
    chip = {'x': 1, 'y': 2, 'acquired': '1980-01-01', 'ubid': 'a/b/c/d'}
    assert chips.identity(chip) == tuple([chip['x'], chip['y'],
                                          chip['ubid'], chip['acquired']])


def test_deduplicate():
    inputs = [{'x': 1, 'y': 2, 'acquired': '1980-01-01', 'ubid': 'a/b/c/d'},
              {'x': 1, 'y': 2, 'acquired': '1980-01-01', 'ubid': 'a/b/c/d'},
              {'x': 2, 'y': 2, 'acquired': '1980-01-01', 'ubid': 'a/b/c/d'},
              {'x': 1, 'y': 3, 'acquired': '1980-01-01', 'ubid': 'a/b/c/d'},
              {'x': 1, 'y': 2, 'acquired': '1980-01-02', 'ubid': 'a/b/c/d'},
              {'x': 1, 'y': 2, 'acquired': '1980-01-01', 'ubid': 'a/b/c'}]

    assert chips.deduplicate(inputs) == tuple(drop(1, inputs))


@test.vcr.use_cassette(test.cassette)
def test_mapped():
    _cfg = cfg.get('chipmunk-ard', env=test.ard_env)
    
    chipmap = chips.mapped(x=test.x,
                           y=test.y,
                           acquired=test.ard_acquired,
                           specmap=specs.mapped(ubids=cfg.ubids.get('chipmunk-ard'),
                                                specs=_cfg.get('registry_fn')()),
                           chips_fn=_cfg.get('chips_fn'))

    assert len(chipmap) > 0
    assert all(map(lambda x: type(x) is dict, concat(chipmap.values())))
    assert len(list(concat(chipmap.values()))) > 0


def test_rsort():
    chipseq = [{'acquired': 3}, {'acquired': 1}, {'acquired': 2}]
    assert [3, 2, 1] == list(map(lambda x: x['acquired'], chips.rsort(chipseq)))



import vcr as _vcr

# chipmunk parameters for test data
x = -2094585
y = 1952805
x_nodata = -183585.0
y_nodata = 302805.0
ard_acquired = '2010/2013'
aux_acquired = '1999/2002'
profile = 'chipmunk-ard'
ard_env = {'CHIPMUNK_URL': 'http://localhost:5656'}
aux_env = {'CHIPMUNK_URL': 'http://localhost:5656'}
cassette = 'test/resources/chipmunk-ard-v1-cassette.yaml'
vcr = _vcr.VCR(record_mode='new_episodes')



from cytoolz import first
from cytoolz import get_in
from cytoolz import partial
from cytoolz import second
from merlin import cfg
from merlin import chips
from merlin import specs
from merlin.formats import pyccd
import test


def data(cfg, acquired):
    x, y = get_in(['chip', 'proj-pt'], cfg['snap_fn'](x=test.x, y=test.y))

    # get specs
    specmap = cfg['specs_fn'](specs=cfg['registry_fn']())

    # get function that will return chipmap.
    # Don't create state with a realized variable to preserve memory
    chipmap = partial(chips.mapped,
                      x=test.x,
                      y=test.y,
                      acquired=acquired,
                      specmap=specmap,
                      chips_fn=cfg['chips_fn'])

    # calculate locations chip.  There's another function
    # here to be split out and organized.
    
    grid = first(filter(lambda x: x['name'] == 'chip',
                        cfg['grid_fn']()))

    cw, ch = specs.refspec(specmap).get('data_shape')
    
    locations = chips.locations(x=x,
                                y=y,
                                cw=cw,
                                ch=ch,
                                rx=grid.get('rx'),
                                ry=grid.get('ry'),
                                sx=grid.get('sx'),
                                sy=grid.get('sy'))
    
    return cfg['format_fn'](x=x,
                            y=y,
                            locations=locations,
                            dates_fn=cfg['dates_fn'],
                            specmap=specmap,
                            chipmap=chipmap(),
                            decode_fn=cfg['decode_fn']) 


@test.vcr.use_cassette(test.cassette)
def test_pyccd():

    d = data(cfg.get('chipmunk-ard', env=test.ard_env),
             test.ard_acquired)

    # we are only testing the structure of the response here.
    # Full data validation is being done in the test for merlin.create()
    assert type(d) is tuple
    assert len(d) == 10000
    assert type(first(d)) is tuple
    assert type(first(first(d))) is tuple
    assert type(second(first(d))) is dict
    assert type(second(second(first(d)))) is tuple or list
    assert len(second(second(first(d)))) > 0


@test.vcr.use_cassette(test.cassette)
def test_aux():

    d = data(cfg.get('chipmunk-aux', env=test.aux_env),
             test.aux_acquired)

    # we are only testing the structure of the response here.
    # Full data validation is being done in the test for merlin.create()
    assert type(d) is tuple
    assert len(d) == 10000
    assert type(first(d)) is tuple
    assert type(first(first(d))) is tuple
    assert type(second(first(d))) is dict
    assert type(second(second(first(d)))) is tuple or list
    assert len(second(second(first(d)))) > 0



from cytoolz import get_in
import functools
import logging
import numpy as np


logger = logging.getLogger(__name__)


def minbox(points):
    """Returns the minimal bounding box necessary to contain points

    Args:
        points (tuple, list, set): ((0,0), (40, 55), (66, 22))

    Returns:
        dict: {ulx, uly, lrx, lry}

    Example:
        >>> minbox((0, 0), (40, 55), (66,22))
        {'ulx': 0, 'uly': 55, 'lrx': 66, 'lry': 0}
    """

    x, y = [point[0] for point in points], [point[1] for point in points]
    return {'ulx': min(x), 'lrx': max(x), 'lry': min(y), 'uly': max(y)}


@functools.singledispatch
def coordinates(points, grid, snap_fn):
    """Returns grid coordinates contained within points.

    Points may be specified as dicts, tuples, lists, or sets.
     
        - dict with keys ulx, uly, lrx, lry
        - sequence of sequences: ((0,0), (100, 167), (-212, 6621))

    Irregular perimeters may be specified in sequences as points will be minboxed.
    
    Points as dicts are an implicit minbox.

    Args:
        points (collection): Points outlining an area
        grid (dict)        : The target grid: {'name': 'chip', 'sx': 3000, 'sy': 3000, 'rx': 1, 'ry': -1}
        snap_fn (func)     : A function that accepts x, y and returns a snapped x, y

    Returns:
        tuple: tuple of tuples of grid coordinates ((x1,y1), (x2,y2) ...)

    Example:
        >>> grid = {'name': 'chip', 'sx': 500, 'sy': 500, 'rx': 1, 'ry': 1}
        >>> sfn = partial(chipmunk.snap, url='http://localhost:5656')
        >>> coordinates({'ulx': -1001, 'uly': 1000, 'lrx': -500, 'lry': 500}, 
                        grid=grid, 
                        snap_fn=sfn)
        ((-3585.0, 2805.0),
         (-3085.0, 2805.0),
         (-2585.0, 2805.0),
         (-2085.0, 2805.0),
         (-1585.0, 2805.0),
         (-1085.0, 2805.0),
         (-585.0, 2805.0))
        
        >>> grid = {'name': 'chip', 'sx': 3000, 'sy': 3000, 'rx': 1, 'ry': -1}
        >>> coordinates(((112, 443), (112, 500), (100, 443)),
                        grid=grid, 
                        snap_fn=sfn})
        ((-585.0, 2805.0),)
    """
    
    logger.warn("coordinates() not implemented for type:{}".format(type(points)))
    return None


@coordinates.register(dict)
def _(points, grid, snap_fn):
    """Returns grid coordinates contained within a dict of points.

    Args:
        points  (dict): {'ulx': 0, 'uly': 15.5, 'lrx': 55.3, 'lry':-1000.12}
        grid    (dict): The target grid: {'name': 'chip', 'sx': 3000, 'sy': 3000, 'rx': 1, 'ry': -1}
        snap_fn (func): A function that accepts x, y and returns a snapped x, y

    Returns:
        tuple: tuple of tuples of grid coordinates ((x1,y1), (x2,y1) ...)

    This example assumes a grid size of 500x500 pixels.

    Example:
        >>> coordinates = coordinates({ulx: -1001, uly: 1000, lrx: -500, lry: 500},
                                      grid={'name': 'chip', 'sx': 500, 'sy': 500, 'rx': 1, 'ry': -1}, 
                                      snap_fn=some_func)

        ((-1000, 500), (-500, 500), (-1000, -500), (-500, -500))
    """
    
    # snap start/end x & y
    start_x, start_y = get_in([grid.get('name'), 'proj-pt'], snap_fn(x=points.get('ulx'), y=points.get('uly')))
    end_x,   end_y   = get_in([grid.get('name'), 'proj-pt'], snap_fn(x=points.get('lrx'), y=points.get('lry')))

    # get x and y scale factors multiplied by reflection
    x_interval = grid.get('sx') * grid.get('rx')
    y_interval = grid.get('sy') * grid.get('ry')
    
    return tuple((x, y) for x in np.arange(start_x, end_x + x_interval, x_interval)
                        for y in np.arange(start_y, end_y + y_interval, y_interval))


@coordinates.register(tuple)
@coordinates.register(list)
@coordinates.register(set)
def _(points, grid, snap_fn):
    """Returns coordinates from a sequence of points.  Performs minbox
    operation on points, thus irregular geometries may be supplied.

    Args:
        points        : a sequence (list, tuple or set) of coordinates.
        grid (dict)   : {'name': 'chip', 'sx': 3000, 'sy': 3000, 'rx': 1, 'ry': -1}
        snap_fn (func): A function that accepts x, y and returns snapped x,y

    Returns:
        tuple: chip coordinates

    Example:
        >>> xys = coordinates(((112, 443), (112, 500), (100, 443)),
                              grid={'sx': 3000, 'sy': 3000}, 
                              snap_fn=some_func})
        >>> ((100, 500),)
    """

    return coordinates(minbox(points), grid=grid, snap_fn=snap_fn)


def extents(ulx, uly, grid):
    """Given an ulx, uly and grid, returns the grid extents.  

    ulx and uly are not translated during this calculation.
 
    Args:
        ulx (float): 0
        uly (float): 0
        grid (dict): {'rx', 'ry', 'sx', 'sy'}

    Returns:
        dict: {'ulx', 'uly', 'lrx', 'lry'}

    Example:
        >>> extents(ulx=0, 
                    uly=0, 
                    grid={'rx': 1, 'ry': -1, 'sx': 3000, 'sy': 3000}
        {'ulx': 0, 'uly': 0, 'lrx': 2999, 'lry': -2999}
    """
    
    return {'ulx': ulx,
            'uly': uly,
            'lrx': ulx + (grid.get('rx') * grid.get('sx')) - int(np.sign(grid.get('rx'))),
            'lry': uly + (grid.get('ry') * grid.get('sy')) - int(np.sign(grid.get('ry')))}



from cytoolz import assoc
from cytoolz import merge
from functools import partial
from merlin import chipmunk
from merlin import chips
from merlin import dates
from merlin import formats
from merlin import specs
import os


ubids = {'chipmunk-ard': {'reds':     ['LC08_SRB4',    'LE07_SRB3',    'LT05_SRB3',    'LT04_SRB3'],
                          'greens':   ['LC08_SRB3',    'LE07_SRB2',    'LT05_SRB2',    'LT04_SRB2'],
                          'blues':    ['LC08_SRB2',    'LE07_SRB1',    'LT05_SRB1',    'LT04_SRB1'],
                          'nirs':     ['LC08_SRB5',    'LE07_SRB4',    'LT05_SRB4',    'LT04_SRB4'],
                          'swir1s':   ['LC08_SRB6',    'LE07_SRB5',    'LT05_SRB5',    'LT04_SRB5'],
                          'swir2s':   ['LC08_SRB7',    'LE07_SRB7',    'LT05_SRB7',    'LT04_SRB7'],
                          'thermals': ['LC08_BTB10',   'LE07_BTB6',    'LT05_BTB6',    'LT04_BTB6'],
                          'qas':      ['LC08_PIXELQA', 'LE07_PIXELQA', 'LT05_PIXELQA', 'LT04_PIXELQA']},

         'chipmunk-aux': {'nlcd':     ['AUX_NLCD'],
                          'nlcdtrn':  ['AUX_NLCDTRN'],
                          'posidex':  ['AUX_POSIDEX'],
                          'mpw':      ['AUX_MPW'],
                          'aspect':   ['AUX_ASPECT'],
                          'slope':    ['AUX_SLOPE'],
                          'dem':      ['AUX_DEM']}
}
         

def profiles(env, profile=None):
    """Retrieve a configuration profile with env applied.

    Args:
        env (dict): Environment variables
        profile (str): Name of profile to load.  If no profile is supplied all profiles
                        are returned.

    Returns:
        dict: Profile or profiles with env substitutions.
    """
    
    __profiles = {
        'chipmunk-ard' : {
            'decode_fn':   chips.decode,
            'grid_fn':     partial(chipmunk.grid,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_GRID_RESOURCE', '/grid')),
            'dates_fn':    dates.symmetric,
            'chips_fn':    partial(chipmunk.chips,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_CHIPS_RESOURCE', '/chips')),
            'specs_fn':    partial(specs.mapped, ubids=ubids['chipmunk-ard']),
            'format_fn':   formats.pyccd,
            'registry_fn': partial(chipmunk.registry,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_REGISTRY_RESOURCE', '/registry')),
            'snap_fn':     partial(chipmunk.snap,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_SNAP_RESOURCE', '/grid/snap')),
            'near_fn':     partial(chipmunk.near,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_NEAR_RESOURCE', '/grid/near'))},
        'chipmunk-aux' : {
            'decode_fn':   chips.decode,
            'grid_fn':     partial(chipmunk.grid,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_GRID_RESOURCE', '/grid')),
            'dates_fn':    dates.single,
            'chips_fn':    partial(chipmunk.chips,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_CHIPS_RESOURCE', '/chips')),
            'specs_fn':    partial(specs.mapped, ubids=ubids['chipmunk-aux']),
            'format_fn':   formats.aux,
            'registry_fn': partial(chipmunk.registry,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_REGISTRY_RESOURCE', '/registry')),
            'snap_fn':     partial(chipmunk.snap,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_SNAP_RESOURCE', '/grid/snap')),
            'near_fn':     partial(chipmunk.near,
                                   url=env.get('CHIPMUNK_URL', None),
                                   resource=env.get('CHIPMUNK_NEAR_RESOURCE', '/grid/near'))},
        }
    
    return __profiles.get(profile, None) if profile else __profiles


def get(profile='chipmunk-ard', env=None):
    """Return a configuration profile.

    Args:
        profile (str): Name of profile.
        env (dict): Environment variables to override os.environ
    Returns:
        dict: A Merlin configuration
    """
    
    p = profiles(env=merge(os.environ, env if env else {}),
                 profile=profile)

    return assoc(p, 'profile', profile)



from base64 import b64decode
from cytoolz import drop
from cytoolz import first
from cytoolz import get_in
from cytoolz import reduce
from cytoolz import unique
from merlin import functions as f
from merlin import specs
from operator import add
import logging
import numpy as np
import requests

logger = logging.getLogger(__name__)


def locations(x, y, cw, ch, rx, ry, sx, sy):
    """Computes locations for array elements that fall within the shape
    specified by chip_spec['data_shape'] using the x & y as the origin.  
    locations() does not snap() x & y... this
    should be done prior to calling locations() if needed.

    Args:
        x: x coordinate
        y: y coordinate
        cw: chip width in pixels (e.g. 100 pixels)
        ch: chip height in pixels (e.g. 100 pixels)
        rx: x reflection (e.g. 1)
        ry: y reflection (e.g. -1)
        sx: x scale (e.g. 3000 meters)
        sy: y scale (e.g. 3000 meters)
    Returns:
        a two (three) dimensional numpy array of [x, y] coordinates
    """

    pw = (sx * rx) / cw # e.g. 30 meters
    ph = (sy * ry) / ch # e.g. -30 meters
    
    # determine ends
    endx = x + cw * pw
    endy = y + ch * ph

    #################################################
    # WARNING: The following line would transpose the
    # resulting matrix by 90 degrees.  In order to
    # generate the proper row major matrix the order
    # y and x are created inside mgrid matters
    #
    # x, y = np.mgrid[x:endx:pw, y:endy:ph]
    #################################################
    
    # build arrays of end - start / step shape
    # flatten into 1d, concatenate and reshape to fit chip
    _y, _x = np.mgrid[y:endy:ph, x:endx:pw]
    matrix = np.c_[_x.ravel(), _y.ravel()]
    return np.reshape(matrix, (cw, ch, 2))


def dates(chips):
    """Dates for a sequence of chips

    Args:
        chips: sequence of chips

    Returns:
        tuple: datestrings
    """

    return tuple([c['acquired'] for c in chips])


def trim(chips, dates):
    """Eliminates chips that are not from the specified dates

    Args:
        chips: Sequence of chips
        dates: Sequence of dates that should be included in result

    Returns:
        tuple: filtered chips
    """

    return tuple(filter(lambda c: c['acquired'] in dates, chips))


def decode(chip):
    """Removes base64 encoding of chip data

    Args:
        chip: A chip

    Returns:
        a decode chip
    """
    
    chip['data'] = b64decode(chip['data'])
    return chip


def chip_to_numpy(chip, chip_spec):
    """Removes base64 encoding of chip data and converts it to a numpy array

    Args:
        chip: A chip
        chip_spec: Corresponding chip_spec

    Returns:
        a decoded chip with data as a shaped numpy array
    """
    
    shape = chip_spec['data_shape']
    dtype = chip_spec['data_type'].lower()
    data  = chip['data']

    chip['data'] = np.frombuffer(data, dtype).reshape(*shape)
    return chip


def to_numpy(chips, spec_index):
    """Converts the data for a sequence of chips to numpy arrays

    Args:
        chips (sequence): a sequence of chips
        spec_index (dict): chip_specs keyed by ubid

    Returns:
        sequence: chips with data as numpy arrays
    """

    return map(lambda c: chip_to_numpy(c, spec_index[c['ubid']]), chips)


def identity(chip):
    """Determine the identity of a chip.

    Args:
        chip (dict): A chip

    Returns:
        tuple: Tuple of the chip identity field
    """

    return tuple([chip['x'], chip['y'],
                  chip['ubid'], chip['acquired']])


def deduplicate(chips):
    """Accepts a sequence of chips and returns a sequence of chips minus
    any duplicates.  A chip is considered a duplicate if it shares an x, y, UBID
    and acquired date with another chip.

    Args:
        chips (sequence): Sequence of chips

    Returns:
        tuple: A nonduplicated tuple of chips
    """

    return tuple(unique(chips, key=identity))


def mapped(x, y, acquired, specmap, chips_fn):
    """Maps chips_fn results to keys from specmap.

    Args:
        x (int): x coordinate
        y (int): y coordinate
        acquired (str): iso8601 date range 
        specmap (dict): map of specs
        chips_fn (fn): function to return chips

    Returns:
        dict: {k: chips_fn()}

    """
    
    return {k: chips_fn(x=x, y=y, acquired=acquired, ubids=specs.ubids(v)) for k, v in specmap.items()}


def rsort(chips, key=lambda c: c['acquired']):
    """Reverse sorts a sequence of chips by date.

    Args:
        chips: sequence of chips

    Returns:
        sorted sequence of chips
    """

    return tuple(f.rsort(chips, key=key))



from cytoolz import first
from cytoolz import reduce
from cytoolz import second
from cytoolz import thread_first
from dateutil import parser
from merlin import chips
from merlin import functions as f
from operator import eq
from operator import or_
import re


def to_ordinal(datestring):
    """Extract an ordinal date from a date string

    Args:
        datestring (str): date value

    Returns:
        int: ordinal date
    """

    return parser.parse(datestring).toordinal()


def startdate(acquired):
    """Returns the startdate from an acquired date string

    Args:
        acquired (str): / separated date range in iso8601 format

    Returns:
        str: Start date
    """

    return acquired.split('/')[0]


def enddate(acquired):
    """Returns the enddate from an acquired date string

    Args:
        acquired (str): / separated date range in iso8601 format

    Returns:
        str: End date
    """

    return acquired.split('/')[1]


def is_acquired(acquired):
    """Is the date string a / separated date range in iso8601 format?

    Args:
        acquired: A date string

    Returns:
        bool: True or False
    """

    # 1980-01-01/2015-12-31
    regex = '^[0-9]{4}-[0-9]{2}-[0-9]{2}\/[0-9]{4}-[0-9]{2}-[0-9]{2}$'
    return bool(re.match(regex, acquired))


def mapped(chipmap):
    """Transform a dict of chips into a dict of datestrings

    Args:
        chipmap (dict): {k: [chips]}

    Returns:
        dict:  {k: [datestring2, datestring1, datestring3]}
    """

    return {k: chips.dates(v) for k, v in chipmap.items()}


def minmax(dates):
    """Returns an iso8601 daterange string that represents the
       min and max datemap.values().

    Args:
        datestrings: [d1, d2, d3,]

    Returns:
        ['min_date/max_date',]

    Example:
        >>> minmax(['2008-01-01', '2010-01-01', '2009-01-01'])
        "2008-01-01/2010-01-01"
    """

    return '{}/{}'.format(min(dates), max(dates)) if dates else '/'


def symmetric(datemap):
    """Returns a sequence of dates that are common to all map values if
    all datemap values are represented, else Exception.
    
    Args:
        datemap: {key: [datestrings,]}

    Returns:
        Sequence of date strings or Exception

    Example:

        >>> symmetric({"reds":  [ds3, ds1, ds2],
                       "blues": [ds2, ds3, ds1]})
        [ds2, ds3, ds1]
        >>>
        >>> symmetric({"reds":  [ds3, ds1],
                       "blues": [ds2, ds3, ds1]})
        Exception: assymetric dates detected - {'reds':[ds3, ds1]} != {'blues':[ds2, ds3, ds1]}
    """

    def check(a, b):
        """Reducer for efficiently comparing two unordered sequences.
        Executes in linear(On) time.

        Args:
            a: {k:[datestring1, datestring2...]}
            b: {k:[datestring2, datestring1...]}

        Returns:
            b if a == b, else Exception with details
        """

        if f.seqeq(second(a), second(b)):
            return b
        else:
            msg = ('assymetric dates detected - {} != {}'
                   .format(first(a), first(b)))
            msga = '{}{}'.format(first(a), second(a))
            msgb = '{}{}'.format(first(b), second(b))
            raise Exception('\n\n'.join([msg, msga, msgb]))

    return second(reduce(check, datemap.items()))


def single(datemap):
    """Returns a sequence of iso8601 daterange strings
       if each datemap.values() is length <= 1, else Exception.

    Args:
        datemap: {key: [datestring,], key2: [datestring,], ...}

    Returns:
        ['ds2', 'ds3', 'ds1'] or Exception

    Example:

        >>> single({"nlcd": [ds3,],
                    "dem":  [ds2,]})
        [ds3, ds2]
        >>>
        >>> single({"nlcd":  [ds3, ds1],
                    "dem":   [ds2,]})
        Exception: assymetric dates detected - {'nlcd':[ds3, ds1], 'dem':[ds2]}
    """
    
    if or_(all(map(lambda a: eq(1, len(a)), datemap.values())),
           all(map(lambda a: eq(0, len(a)), datemap.values()))):
        return list(f.flatten(datemap.values()))
    else:
        raise Exception('assymetric dates detected - {}'.format(datemap))
    

def rsort(dateseq):
    """ Reverse sorts a sequence of dates.

    Args:
        dateseq: sequence of dates

    Returns:
        sequence: reverse sorted sequence of dates
    """
    
    return f.rsort(dateseq)



"""functions.py is a module of generalized, reusable functions"""

from collections import Counter
from cytoolz import drop
from cytoolz import first
from cytoolz import get_in
from cytoolz import merge
from cytoolz import partial
from cytoolz import reduce
from cytoolz import second
from datetime import datetime
from functools import singledispatch
import functools
import hashlib
import itertools
import json
import logging
import numpy as np
import re

logger = logging.getLogger(__name__)


def extract(sequence, elements):
    """Given a sequence (possibly with nested sequences), extract
    the element identifed by the elements sequence.

    Args:
        sequence: A sequence of elements which may be other sequences
        elements: Sequence of nested element indicies (in sequence parameter)
            to extract

    Returns:
        The target element

    Example:
        >>> inputs = [1, (2, 3, (4, 5)), 6]
        >>> extract(inputs, [0])
        >>> 1
        >>> extract(inputs, [1])
        >>> (2, 3, (4, 5))
        >>> extract(inputs, [1, 0])
        >>> 2
        >>> extract(inputs, [1, 1])
        >>> 3
        >>> extract(inputs, [1, 2])
        >>> (4, 5)
        >>> extract(inputs, [1, 2, 0])
        >>> 4
    ...
    """

    e = tuple(elements)
    if len(e) == 0 or not getattr(sequence, '__iter__', False):
        return sequence
    else:
        seq = sequence[first(e)]
        return extract(seq, drop(1, e))


def flatten(iterable):
    """Reduce dimensionality of iterable containing iterables

    Args:
        iterable: A multi-dimensional iterable

    Returns:
        A one dimensional iterable
    """

    return itertools.chain.from_iterable(iterable)


def intersection(items):
    """Returns the intersecting set contained in items

    Args:
        items: Two dimensional sequence of items

    Returns:
        Intersecting set of items

    Example:
        >>> items = [[1, 2, 3], [2, 3, 4], [3, 4, 5]]
        >>> intersection(items)
        {3}
    """

    return set.intersection(*(map(lambda x: set(x), items)))


def sha256(string):
    """Computes and returns a sha256 digest of the supplied string

    Args:
        string (str): string to digest

    Returns:
        digest value
    """

    return hashlib.sha256(string.encode('UTF-8')).hexdigest()


def md5(string):
    """Computes and returns an md5 digest of the supplied string

    Args:
        string: string to digest

    Returns:
        digest value
    """

    return hashlib.md5(string.encode('UTF-8')).hexdigest()


def sort(iterable, key=None):
    """Sorts an iterable"""

    return sorted(iterable, key=key, reverse=False)


def rsort(iterable, key=None):
    """Reverse sorts an iterable"""

    return sorted(iterable, key=key, reverse=True)


@singledispatch
def serialize(arg):
    """Converts datastructure to json, computes digest

    Args:
        dictionary: A python dict

    Returns:
        tuple: (digest,json)
    """

    s = json.dumps(arg, sort_keys=True, separators=(',', ':'),
                   ensure_ascii=True)
    return md5(s), s


@serialize.register(set)
def _(arg):
    """Converts set to list then serializes the resulting value"""

    return serialize(sorted(list(arg)))


def deserialize(string):
    """Reconstitues datastructure from a string.

    Args:
        string: A serialized data structure

    Returns:
        Data structure represented by string parameter
    """

    return json.loads(string)


def flip_keys(dods):
    """Accepts a dictionary of dictionaries and flips the outer and inner keys.
    All inner dictionaries must have a consistent set of keys or key Exception
    is raised.

    Args:
        dods: dict of dicts

    Returns:
        dict of dicts with inner and outer keys flipped

    Example:
        >>> dods = {"reds":   {(0, 0): [110, 110, 234, 664],
                               (0, 1): [23, 887, 110, 111]},
                    "greens": {(0, 0): [120, 112, 224, 624],
                               (0, 1): [33, 387, 310, 511]},
                    "blues":  {(0, 0): [128, 412, 244, 654],
                               (0, 1): [73, 987, 119, 191]},
        >>> flip_keys(dods)
        {(0, 0): {"reds":   [110, 110, 234, 664],
                  "greens": [120, 112, 224, 624],
                  "blues":  [128, 412, 244, 654], ... },
         (0, 1), {"reds":   [23, 887, 110, 111],
                  "greens": [33, 387, 310, 511],
                  "blues":  [73, 987, 119, 191], ... }}

    """

    def flip(innerkeys, outerkeys, inputs):
        for ik in innerkeys:
            yield({ik: {ok: inputs[ok][ik] for ok in outerkeys}})

    outerkeys = set(dods.keys())
    innerkeys = set(reduce(lambda accum, v: accum + v,
                           [list(dods[ok].keys()) for ok in outerkeys]))
    return merge(flip(innerkeys, outerkeys, dods))


def cqlstr(string):
    """Makes a string safe to use in Cassandra CQL commands

    Args:
        string: The string to use in CQL

    Returns:
        str: A safe string replacement
    """

    return re.sub('[-:.]', '_', string)


def represent(item):
    """Represents callables and values consistently

    Args:
        item: The item to represent

    Returns:
        Item representation
    """

    return repr(item.__name__) if callable(item) else repr(item)


def isnumeric(value):
    """Does a string value represent a number (positive or negative?)

    Args:
        value (str): A string

    Returns:
        bool: True or False
    """

    try:
        float(value)
        return True
    except:
        return False


def timed(f):
    """Timing wrapper for functions.  Prints start and stop time to INFO
    along with function name, arguments and keyword arguments.

    Args:
        f (function): Function to be timed

    Returns:
        function: Wrapped function
    """

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        """Wrapper function."""
        msg = '{}(args={}, kwargs={}):'.format(f.__name__,  args, kwargs)
        logger.info('{} start:{}'.format(msg, datetime.now().isoformat()))
        r = f(*args, **kwargs)
        logger.info('{} stop:{}'.format(msg, datetime.now().isoformat()))
        return r
    return wrapper


def seqeq(a, b):
    """Determine if two unordered sequences are equal.

    Args:
        a: sequence a
        b: sequence b

    Returns:
        bool: True or False
    """

    #runs in linear (On) time.
    return Counter(a) == Counter(b)


def issubset(a, b):
    """Determines if a exists in b

    Args:
        a: sequence a
        b: sequence b

    Returns:
        bool: True or False
    """

    return set(a).issubset(set(b))


def difference(a, b):
    """Subtracts items in b from items in a.

    Args:
        a: sequence a
        b: sequence b

    Returns:
        set: items that exist in a but not b
    """

    return set(a) - set(b)


def chexists(dictionary, keys, check_fn):
    """applies check_fn against dictionary minus keys then ensures the items
    returned from check_fn exist in dictionary[keys]

    Args:
        dictionary (dict): {key: [v1, v3, v2]}
        keys (sequence): A sequence of keys in dictionary
        check_fn (function): Function that accepts dict and returns
                             sequence of items or Exception

    Returns:
        A sequence of items that are returned from check_fn and exist in
        dictionary[keys] or Exception
    """

    def exists_in(superset, subset):
        if issubset(subset, second(superset)):
            return True
        else:
            msg =  '{} is missing data.'.format(first(superset))
            msg2 = '{} is not a subset of {}'.format(subset, second(superset))
            raise Exception('\n\n'.join([msg, msg2]))

    popped  = {k: dictionary[k] for k in keys}
    checked = check_fn({k: dictionary[k] for k in difference(dictionary, keys)})
    all(map(partial(exists_in, subset=checked), popped.items()))
    return checked


def insert_into_every(dods, key, value):
    """Insert key:values into every subdictionary of dods.

    Args:
        dods: dictionary of dictionaries
        key: key to hold values in subdictionaires
        value: value to associate with key

    Returns:
        dict: dictionary of dictionaries with key:values inserted into each
    """

    def update(d, v):
        d.update({key: v})
        return d

    return {k: update(v, value) for k, v in dods.items()}


@singledispatch
def denumpify(arg):
    """Converts numpy datatypes to python datatypes

    bool_ and bool8 are converted to Python bool
    float64, float32 and float16's are converted to Python float()
    intc, intp, int8, int16, int32, int64, uint8, uint16, uint32 and uint64 are converted to Python int()
    complex_, complex64 and complex128 are converted to Python complex()
    None is returned as None
    numpy ndarrays are returned as list()    

    Python lists, maps, sets and tuples are returned with all values converted
    to Python types (recursively).

    If there is no implemented converter, returns arg.

    Args:
        arg: A (possibly numpy) data structure

    Returns:
        A Python data structure
    """
    
    return arg


@denumpify.register(np.bool_)
@denumpify.register(np.bool8)
def _(arg):
    """Converts numpy bools to python bools"""
    return bool(arg)


@denumpify.register(np.float64)
@denumpify.register(np.float32)
@denumpify.register(np.float16)
def _(arg):
    """Converts numpy floats to python floats"""
    return float(arg)


@denumpify.register(np.intc)
@denumpify.register(np.intp)
@denumpify.register(np.int8)
@denumpify.register(np.int16)
@denumpify.register(np.int32)
@denumpify.register(np.int64)
@denumpify.register(np.uint8)
@denumpify.register(np.uint16)
@denumpify.register(np.uint32)
@denumpify.register(np.uint64)
def _(arg):
    """Converts numpy ints to python ints"""
    return int(arg)


@denumpify.register(np.complex_)
@denumpify.register(np.complex64)
@denumpify.register(np.complex128)
def _(arg):
    """Converts numpy complex numbers to python complex"""
    return complex(arg)


@denumpify.register(np.ndarray)
def _(arg):
    """Converts ndarray to listset to list"""
    return arg.tolist()


@denumpify.register(list)
def _(arg):
    """Converts list values to Python types"""
    return [denumpify(l) for l in arg]


@denumpify.register(set)
def _(arg):
    """Converts set values to Python types"""
    return set(denumpify(list(arg)))


@denumpify.register(tuple)
def _(arg):
    """Converts tuple values to Python types"""
    return tuple(denumpify(list(arg)))

        
@denumpify.register(dict)
def _(arg):
    """Converts dict values to Python types"""
    return {k: denumpify(v) for k, v in arg.items()}


from cytoolz import first
from cytoolz import filter
from merlin import cfg
from merlin import geometry as g
import test


def chip_grid(config):
    return first(filter(lambda x: x['name'] == 'chip', config.get('grid_fn')()))


def tile_grid(config):
    return first(filter(lambda x: x['name'] == 'tile', config.get('grid_fn')()))


def test_minbox():
    one = {'ulx': 0, 'uly': 0, 'lrx': 0, 'lry': 0}
    assert g.minbox(((0,0),)) == one

    two = {'ulx': 0, 'uly': 10, 'lrx': 10, 'lry': 0}
    assert g.minbox(((0,0), (10, 10))) == two

    three = {'ulx': -50, 'uly': 3, 'lrx': 10, 'lry': -8}
    assert g.minbox(((-50,0), (10, 3), (5, -8))) == three

    four = {'ulx': -5, 'uly': 33, 'lrx': 111, 'lry': -66}
    assert g.minbox(((0, 11), (-5, -5), (3, 33), (111, -66))) == four

    five = {'ulx': -5 , 'uly': 5 , 'lrx': 4 , 'lry': -4 }
    assert g.minbox(((1, 1), (2, 2), (-3, -3), (4, -4), (-5, 5))) == five


@test.vcr.use_cassette(test.cassette)
def test_coordinates():
    _cfg = cfg.get('chipmunk-ard', env=test.ard_env)
    grid = chip_grid(_cfg)
    sfn  = _cfg.get('snap_fn')
    
    expected = ((-585.0, 2805.0), (-585.0, -195.0), (2415.0, 2805.0), (2415.0, -195.0))
    result   = g.coordinates({'ulx': 0, 'uly': 0, 'lrx': 3000, 'lry': -3000}, grid=grid, snap_fn=sfn)
    assert expected == result

    grid     = tile_grid(_cfg)
    expected = ((-15585.0, 14805.0),)
    result = g.coordinates({'ulx': 0, 'uly': 0, 'lrx': 3000, 'lry': -3000}, grid=grid, snap_fn=sfn)
    assert expected == result

    grid     = chip_grid(_cfg)
    expected = ((-3585.0, 5805.0), (-3585.0, 2805.0), (-585.0, 5805.0), (-585.0, 2805.0))
    result   = g.coordinates(((0, 0), (-590, 0), (0, 2806)), grid=grid, snap_fn=sfn)
    assert expected == result

    grid     = chip_grid(_cfg)
    expected = ((-3585.0, 5805.0), (-3585.0, 2805.0), (-585.0, 5805.0), (-585.0, 2805.0))
    result   = g.coordinates([(0, 0), (-590, 0), (0, 2806)], grid=grid, snap_fn=sfn)
    assert expected == result

    grid     = chip_grid(_cfg)
    expected = ((-3585.0, 5805.0), (-3585.0, 2805.0), (-585.0, 5805.0), (-585.0, 2805.0))
    result   = g.coordinates(set(((0, 0), (-590, 0), (0, 2806))), grid=grid, snap_fn=sfn)
    assert expected == result


def tests_extents():
    results = g.extents(ulx=0, uly=0, grid={'rx': 1, 'ry': -1, 'sx': 3000, 'sy': 3000})
    assert results == {'ulx': 0, 'uly': 0, 'lrx': 2999, 'lry': -2999}



from merlin import dates
import pytest

def test_to_ordinal():
    assert dates.to_ordinal('1999-01-01') == 729755
    assert dates.to_ordinal('1999/11/01') == 730059


def test_startdate():
    assert dates.startdate('1980-01-01/1982-01-01') == '1980-01-01'


def test_enddate():
    assert dates.enddate('1980-01-01/1982-01-01') == '1982-01-01'


def test_is_acquired():
    assert dates.is_acquired('1980-01-01/1982-01-01') is True
    assert dates.is_acquired('1980-01-011982-01-01') is False


def test_minmax():
    datestrs = ['1980', '1979', '1990']

    assert '1979/1990' == dates.minmax(datestrs)

    
def test_symmetric():
    datemap = {'a': [1, 2, 3],
               'b': [2, 1, 3],
               'c': [3, 2, 1]}

    assert set([1, 2, 3]) == set(dates.symmetric(datemap))

    with pytest.raises(Exception):

        datemap = {'a': [1, 2, 3],
                   'b': [2, 1, 3],
                   'c': [3]}

        dates.symmetric(datemap)


def test_single():

    # single date from each key-value should pass
    datemap = {'a': [1,],
               'b': [2,],
               'c': [3,]}

    assert set([1, 2, 3]) == set(dates.single(datemap))

    # zero dates from each is also acceptable
    datemap = {'a': [],
               'b': [],
               'c': []}

    assert set([]) == set(dates.single(datemap))

    # more than 1 date is an error
    with pytest.raises(Exception):

        datemap = {'a': [1, 2, 3],
                   'b': [2, 1, 3],
                   'c': [3, 2, 1]}

        dates.single(datemap)

    # more than 1, misshappen is an Exception
    with pytest.raises(Exception):

        datemap = {'a': [1, 2, 3],
                   'b': [2, 1, 3],
                   'c': [3]}

        dates.single(datemap)

        
    # assymetric dates are an Exception
    with pytest.raises(Exception):

        datemap = {'a': [1,],
                   'b': [2,],
                   'c': []}

        dates.single(datemap)

    
def test_rsort():
    assert [3, 2, 1] == dates.rsort([3, 1, 2])
 



from setuptools import find_packages
from setuptools import setup
import os

def readme():
    with open('README.rst') as f:
        return f.read()

def version():
    return '2.3.4-rc1'

setup(name='lcmap-merlin',
      version=version(),
      description='Python client library for LCMAP rasters',
      long_description=readme(),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: Public Domain',
        'Programming Language :: Python :: 3',
      ],
      keywords='usgs lcmap eros',
      url='http://github.com/usgs-eros/lcmap-merlin',
      author='USGS EROS LCMAP',
      author_email='',
      license='Unlicense',
      packages=find_packages(),
      install_requires=[
          'cytoolz',
          'numpy',
          'requests',
          'python-dateutil',
      ],
      # List additional groups of dependencies here (e.g. development
      # dependencies). You can install these using the following syntax,
      # for example:
      # $ pip install -e .[test]
      extras_require={
          'test': ['pytest',
                   'pytest-cov',
                   'hypothesis',
                   'vcrpy',
                  ],
          'doc': ['sphinx',
                  'sphinx-autobuild',
                  'sphinx_rtd_theme'],
          'dev': ['twine'],
      },
      # entry_points={
          #'console_scripts': [''],
      # },
      include_package_data=True,
      zip_safe=False)



from cytoolz import cons
from cytoolz import merge
from merlin import specs


def test_only():
    registry = [{'key': 1, 'ubid': 'a'}, {'key': 2, 'ubid': 'b'},
                {'key': 3, 'ubid': 'c'}, {'key': 4, 'ubid': 'd'}]

    expected = tuple([{'key': 1, 'ubid': 'a'}, {'key': 4, 'ubid': 'd'}])
    result   = specs.only(ubids=['a', 'd'], specs=registry)
    assert expected == result

    expected = tuple([{'key': 1, 'ubid': 'a'}, {'key': 4, 'ubid': 'd'}])
    result   = specs.only(ubids=['d', 'a'], specs=registry)
    assert expected == result
    
    expected = tuple()
    result = specs.only(ubids=['z'], specs=registry)
    assert expected == result


def test_mapped():
    registry = [{'key': 1, 'ubid': 'a'}, {'key': 2, 'ubid': 'b'},
                {'key': 3, 'ubid': 'c'}, {'key': 4, 'ubid': 'd'}]

    ubids = {'red': ['a', 'd'], 'blue': ['b', 'c']}

    expected = {'red':  ({'key': 1, 'ubid': 'a'}, {'key': 4, 'ubid': 'd'}),
                'blue': ({'key': 2, 'ubid': 'b'}, {'key': 3, 'ubid': 'c'})}
                
    result = specs.mapped(ubids=ubids, specs=registry)
    assert expected == result


def test_exist():
    registry = [{'key': 1, 'ubid': 'a'}, {'key': 2, 'ubid': 'b'},
                {'key': 3, 'ubid': 'c'}, {'key': 4, 'ubid': 'd'}]

    ubids = ['a', 'd']
    assert True==specs.exist(ubids=ubids, specs=registry)

    ubids = ['a', 'b', 'x']
    assert False==specs.exist(ubids=ubids, specs=registry)

    
def test_index():
    registry = [{'key': 1, 'ubid': 'a'}, {'key': 2, 'ubid': 'b'},
                {'key': 3, 'ubid': 'c'}, {'key': 4, 'ubid': 'd'}]
    
    results = specs.index(registry)

    # check that dicts were rekeyed into a new dict
    assert all(map(lambda r: r in results, ['a', 'b', 'c', 'd']))

    # check structure of new dict values
    assert all(map(lambda r: 'ubid' in r and 'key' in r, results.values()))


def test_ubids():
    registry = [{'key': 1, 'ubid': 'a'}, {'key': 2, 'ubid': 'b'},
                {'key': 3, 'ubid': 'c'}, {'key': 4, 'ubid': 'd'}]

    data = list(cons({'nope': 'z'}, registry))

    good = filter(lambda f: 'ubid' in f, data)
    
    assert set(map(lambda u: u['ubid'], good)) == set(specs.ubids(data))


def test_ubids():
    ubids = [{'ubid': 1}, {'ubid': 2}, {'ubid': 3}, {'ubid': 4}]
    assert len(specs.ubids(ubids)) == 4


def test_refspec():
    specmap = {'red':  [{'key': 1, 'ubid': 'a'}, {'key': 2, 'ubid': 'b'}],
               'blue': [{'key': 3, 'ubid': 'c'}, {'key': 4, 'ubid': 'd'}]}

    assert type(specs.refspec(specmap)) is dict



from cytoolz import first
from cytoolz import last
from cytoolz import partial
from cytoolz import sliding_window
from merlin import cfg
from merlin import chips
from merlin import create
from merlin import specs
from operator import eq
from operator import gt
import numpy as np
import test

@test.vcr.use_cassette(test.cassette)    
def test_create_nodata():
    _cfg = cfg.get(profile=test.profile, env=test.ard_env)
    data = create(x=test.x_nodata, y=test.y_nodata, acquired=test.ard_acquired, cfg=_cfg)
    assert len(data) == 0
    assert type(data) is tuple

    
@test.vcr.use_cassette(test.cassette)    
def test_create():
    # data should be shaped: ( ((chip_x, chip_y, x1, y1),{}),
    #                          ((chip_x, chip_y, x1, y2),{}), )

    _cfg = cfg.get(profile=test.profile, env=test.ard_env)    
    data = create(x=test.x, y=test.y, acquired=test.ard_acquired, cfg=_cfg)

    # make sure we have 10000 results and the response structure
    assert len(data) == 10000
    assert isinstance(data, tuple)
    assert isinstance(data[0], tuple)
    assert isinstance(data[0][0], tuple)
    assert isinstance(data[0][1], dict)

    # check key length: chip_x, chip_y, x, y.
    assert len(data[0][0]) == 4

    # check to make sure we have equal length values and that the values
    # are not empty.  FYI -- only spot checking the first returned result
    lens = [len(data[0][1][key]) for key in data[0][1]]
    assert all([eq(*x) for x in sliding_window(2, lens)]) == True

    # make sure everything isn't zero length
    assert all([gt(x, 0) for x in lens]) == True


@test.vcr.use_cassette(test.cassette)
def test_compare_timeseries_to_chip():
    # Make sure the timeseries values match the most recent chip for a spectra.
    # This will validate the chip was not transposed during timeseries creation.
    
    _cfg =   cfg.get(profile=test.profile, env=test.ard_env)
    _ubids = cfg.ubids.get(test.profile).get('reds')
    _chips = _cfg.get('chips_fn')(x=test.x, y=test.y, acquired=test.ard_acquired, ubids=_ubids)
    _index = specs.index(_cfg.get('registry_fn')())
     
    most_recent_chip = last(sorted(_chips, key=lambda d: d['acquired']))
    most_recent_chip = chips.decode(most_recent_chip)
    most_recent_chip = first(chips.to_numpy(spec_index=_index, chips=[most_recent_chip]))

    time_series = create(x=test.x, y=test.y, acquired=test.ard_acquired, cfg=_cfg)
                         
    # this is a 2d 100x100 array of values arranged spatially
    chip_data = most_recent_chip['data']

    # reconstruct a 2d 100x100 array of values from the most recent observation in the time series.  Idea here is it
    # should match the chip_data
    #.reshape(100, 100)
    def observation(record):
        # return x, y, value for the most recent observation in the time series
        return {'x': record[0][2], 'y': record[0][3], 'v': record[1]['reds'][0]}
        
    def lookups(values):
        # need to know the array position of each value, be able to look them up
        # this is a coordinate space to array space mapping
        return  {v:i for i,v in enumerate(values)}

    def build_array(obs, x_lookups, y_lookups):
        # takes all the observations and the x/y lookup tables, returns an ndarray (float)

        # initialze an ndarray that matches the dimensions of x and y
        ar = np.zeros([len(x_lookups), len(y_lookups)], dtype=np.float)

        # populate the ndarray
        for o in obs:
            x = x_lookups.get(o['x'])
            y = y_lookups.get(o['y'])
            ar.itemset((y, x), o.get('v'))
            
        return ar
       
    observations = list(map(observation, time_series))
    xset = sorted(list(set(map(lambda a: a['x'], observations))))
    yset = sorted(list(set(map(lambda a: a['y'], observations))), reverse=True)
    suspect_data = build_array(obs=observations,
                               x_lookups=lookups(xset),
                               y_lookups=lookups(yset))

    assert np.array_equal(suspect_data, chip_data)



from cytoolz import filter
from cytoolz import first
from cytoolz import get_in
from cytoolz import partial
from merlin import cfg
from merlin import chipmunk
from merlin import chips
from merlin import dates
from merlin import files
from merlin import formats
from merlin import functions
from merlin import rods
from merlin import specs


def create(x, y, acquired, cfg):
    """Create a timeseries.

    Args:
        x (int): x coordinate
        y (int): y coordinate
        acquired (string): iso8601 date range
        cfg (dict): A Merlin configuration

    Returns:
        tuple - Results of format_fn applied to results of chips_fn
    """
    
    x, y = get_in(['chip', 'proj-pt'], cfg['snap_fn'](x=x, y=y))

    # get specs
    specmap = cfg['specs_fn'](specs=cfg['registry_fn']())

    # get function that will return chipmap.
    # Don't create state with a realized variable to preserve memory
    chipmap = partial(chips.mapped,
                      x=x,
                      y=y,
                      acquired=acquired,
                      specmap=specmap,
                      chips_fn=cfg['chips_fn'])

    # calculate locations chip.  There's another function
    # here to be split out and organized.
    
    grid = first(filter(lambda x: x['name'] == 'chip',
                        cfg['grid_fn']()))

    cw, ch = specs.refspec(specmap).get('data_shape')
    
    locations = partial(chips.locations,
                        x=x,
                        y=y,
                        cw=cw,
                        ch=ch,
                        rx=grid.get('rx'),
                        ry=grid.get('ry'),
                        sx=grid.get('sx'),
                        sy=grid.get('sy'))
    
    return cfg['format_fn'](x=x,
                            y=y,
                            locations=locations(),
                            dates_fn=cfg['dates_fn'],
                            specmap=specmap,
                            chipmap=chipmap(),
                            decode_fn=cfg['decode_fn']) 



from cytoolz import first
from cytoolz import second
from merlin import cfg
from merlin import functions as f
import numpy as np
import pytest
import test


def test_extract():
    inputs = [1, (2, 3, (4, 5)), 6]
    assert f.extract(inputs, [0]) == 1
    assert f.extract(inputs, [1]) == (2, 3, (4, 5))
    assert f.extract(inputs, [1, 0]) == 2
    assert f.extract(inputs, [1, 1]) == 3
    assert f.extract(inputs, [1, 2]) == (4, 5)
    assert f.extract(inputs, [1, 2, 0]) == 4


def test_flatten():
    result = [1, 2, 3, 4, 5, [6, 7]]
    assert list(f.flatten([[1, 2], [3, 4], [5, [6, 7]]])) == result


def test_intersection():
    items = [[1, 2, 3], [2, 3, 4], [3, 4, 5]]
    assert f.intersection(items) == {3}


def test_sha256():
    assert type(f.sha256('kowalski')) is str


def test_md5():
    assert type(f.md5('kowalski')) is str


def test_sort():
    assert f.sort([2, 1, 3]) == [1, 2, 3]


def test_rsort():
    assert f.rsort([2, 1, 3]) == [3, 2, 1]


def test_serialization():

    def serialize_deserialize(t):
        assert f.deserialize(second(f.serialize(t))) == t

    targets = ['farva', 1, 2.0, True, None]

    all(map(serialize_deserialize, targets))


def test_flip_keys():
    dods = {'a': {1: 'a1', 2: 'a2'},
            'b': {1: 'b1', 2: 'b2'},
            'c': {1: 'c1', 2: 'c2'}}

    result = {1: {'a': 'a1', 'b': 'b1', 'c': 'c1'},
              2: {'a': 'a2', 'b': 'b2', 'c': 'c2'}}

    assert f.flip_keys(dods) == result


def test_cqlstr():
    assert f.cqlstr('-:.') == '___'


def test_represent():

    def ramrod():
        pass

    assert f.represent(ramrod) == "'ramrod'"
    assert f.represent('ramrod') == "'ramrod'"
    assert f.represent(1) == '1'
    assert f.represent(None) == 'None'


def test_isnumeric():
    good = ['1', '-1', '0', '1.0', '-1.0']
    assert all(map(f.isnumeric, good))

    bad = ['a', 'a1', '1a']
    assert not any(map(f.isnumeric, bad))


def test_seqeq():
    assert f.seqeq([1, 2, 3], [2, 3, 1]) is True
    assert f.seqeq({1, 2, 3}, [2, 3, 1]) is True
    assert f.seqeq(tuple([3, 2, 1]), [1, 2, 3]) is True
    assert f.seqeq([1,], [1, 2, 3]) is False
    assert f.seqeq([1, 2, 3], [1,]) is False


def test_issubset():
    assert f.issubset([1, 2], [3, 1, 2]) is True
    assert f.issubset(tuple([2, 1]), [3, 2, 1]) is True
    assert f.issubset({'a': 1, 'b': 2}, {'a': 1, 'c': 3, 'b': 2}) is True
    assert f.issubset([3, 2, 1], [3, 2]) is False


def test_difference():
    a = [1, 2, 3]
    b = [1, 2]
    assert f.difference(a, b) == {3}
    assert f.difference(b, a) == set()


def test_chexists():

    # simple check function that will return the first list
    def check_fn(dictionary):
        return dictionary[first(dictionary)]

    # simulate assymetric list values, telling chexists 'c' is supposed to be
    # bigger.
    d = {'a': [1, 2, 3], 'b': [1, 2, 3], 'c': [1, 2, 3, 4, 5]}
    assert f.chexists(d, ['c'], check_fn) == d['a']

    # make sure we throw an exception if c does not contain all values from
    # check_fn
    with pytest.raises(Exception):
        d = {'a': [1, 2, 3], 'b': [1, 2, 3], 'c': [1, 2, 4, 5]}
        assert f.chexists(d, ['c'], check_fn) == d['a']


def test_insert_into_every():
    dod = {1: {1: 'one'},
           2: {1: 'one'}}

    result = f.insert_into_every(dod, key='newkey', value='newval')

    #assert all(['newkey' in dod.get(key) for key in dod.keys()])
    assert all([dod.get(key).get('newkey') == 'newval' for key in dod.keys()])


def test_denumpify():
    bools = [np.bool_, np.bool8, bool]
    
    ints = [np.intc, np.intp, np.int8, np.int16, np.int32, np.int64,
            np.uint8, np.uint16, np.uint32, np.uint64, int]

    floats = [np.float64, np.float32, np.float16, float]

    complexes = [np.complex_, np.complex64, np.complex128, complex]

    assert all([type(f.denumpify(_b(10))) == bool for _b in bools])
    assert all([type(f.denumpify(_i(10))) == int for _i in ints])
    assert all([type(f.denumpify(_f(10))) == float for _f in floats])
    assert all([type(f.denumpify(_c(10))) == complex for _c in complexes])
    assert type(f.denumpify(np.ndarray([1, 2, 3])) == list)

    assert f.denumpify(None) is None
    assert type(f.denumpify(set([1, 2, 3]))) == set
    assert type(f.denumpify([1, 2, 3])) == list
    assert type(f.denumpify(dict(a=1))) == dict
    assert type(f.denumpify(tuple([1, 2, 3]))) == tuple

    # make sure it's converting subelement correctly
    l = f.denumpify([np.int16(1), np.int16(20), np.int16(3)])
    assert all([type(ll) == int for ll in l])                    



from functools import partial
from cytoolz import excepts
from cytoolz import thread_last
from merlin import chips
import numpy as np

def from_chips(chips):
    """Accepts sequences of chips and returns
    time series pixel rods organized by x, y, t for all chips.
    Chips should be sorted as desired before calling rods() as outputs
    preserve input order.

    Args:
        chips: sequence of chips with data as numpy arrays

    Returns:
        3d numpy array organized by x, y, and t.  Output shape matches
        input chip shape with the chip value replaced by
        another numpy array of chip time series values

    1. For each chip add data to master numpy array.
    2. Transpose the master array
    3. Horizontally stack the transposed master array elements
    4. Reshape the master array to match incoming chip dimensions
    5. Pixel rods are now organized for timeseries access by x, y, t

    >>> chip_one   = np.int_([[11, 12, 13],
                              [14, 15, 16],
                              [17, 18, 19]])
    >>> chip_two   = np.int_([[21, 22, 23],
                              [24, 25, 26],
                              [27, 28, 29]])
    >>> chip_three  = np.int_([[31, 32, 33],
                               [34, 35, 36],
                               [37, 38, 39]])
    >>> master = np.conj([chip_one, chip_two, chip_three])
    >>> np.hstack(master.T).reshape(3, 3, -1)
    array([[[ 11, 21, 31], [ 12, 22, 32], [ 13, 23, 33]],
           [[ 14, 24, 34], [ 15, 25, 35], [ 16, 26, 36]],
           [[ 17, 27, 37], [ 18, 28, 38], [ 19, 29, 39]]])
    """
    
    master = np.conj([c['data'] for c in chips])
    return np.hstack(master.T).reshape(*master[0].shape, -1)


def locate(rods, locations):
    """Combines location information with pixel rods.

    Args:
        rods: Chip shaped numpy array of rods
        locations:  Chip shaped numpy array of locations

    Returns:
        dict: (location):rod for each location and rod in the arrays.

    Incoming locations as 3d array:

    >>> array([[[0,0], [0,1], [0,2]],
               [[1,0], [1,1], [1,2]],
               [[2,0], [2,1], [2,2]]])

    Incoming rods also as 3d array:

    >>> array([[[110,110,234,664], [23,887,110,111], [110,464,223,112]],
               [[111,887,1,110],   [33,111,12,111],  [0,111,66,112]],
               [[12,99,112,110],   [112,87,231,111], [112,45,47,112]]])

    locrods converts locations to:

    >>> locations.reshape(locations.shape[0] * locations.shape[1], -1)
    array([[0, 0],
           [0, 1],
           [0, 2],
           [1, 0],
           [1, 1],
           [1, 2],
           [2, 0],
           [2, 1],
           [2, 2]])

    And rods to:

    >>> rods.reshape(rods.shape[0] * rods.shape[1], -1)
    array([[110, 110, 234, 664],
           [23,  887, 110, 111],
           [110, 464, 223, 112],
           [111, 887, 1,   110],
           [33,  111, 12,  111],
           [0,   111, 66,  112],
           [12,  99,  112, 110],
           [112, 87,  231, 111],
           [112, 45,  47,  112]])

    Then the locations and rods are zipped together via a dictionary
    comprehension and returned.

    >>> {
         (0,0): [110, 110, 234, 664],
         (0,1): [23,  887, 110, 111],
         (0,2): [110, 464, 223, 112],
         (1,0): [111, 887, 1,   110],
         (1,1): [33,  111, 12,  111],
         (1,2): [0,   111, 66,  112],
         (2,0): [12,  99,  112, 110],
         (2,1): [112, 87,  231, 111],
         (2,2): [112, 45,  47,  112]
        }
    """

    flat_locs = locations.reshape(locations.shape[0] * locations.shape[1], -1)
    flat_rods = rods.reshape(rods.shape[0] * rods.shape[1], -1)
    return {tuple(k): v for k, v in zip(flat_locs, flat_rods)}


def identify(rod, x, y):
    """Adds chip ids (chip_x, chip_y) to the key for a rod

    Args:
        rod: dict of (x, y): [values]
        x: x coordinate that identifies the source chip
        y: y coordinate that identifies the source chip
        
    Returns:
        dict: {(chip_x, chip_y, x, y): [values]}
    """

    return {(x, y, k[0], k[1]): v for k, v in rod.items()}


def create(x, y, chipseq, dateseq, locations, spec_index, decode_fn):
    """Transforms a sequence of chips into a sequence of rods
       filtered by date, deduplicated, sorted, located and identified.

       Args:
           x (int): x projection coordinate of chip
           y (int): y projection coordinate of chip
           chipseq (seq): sequence of chips
           dates (seq): sequence of dates that should be included in the rods
           locations (numpy.Array): 2d numpy array of pixel coordinates
           spec_index (dict): specs indexed by ubid
           decode_fn (fn): Function that accepts a chip and returns a decoded chip
                           (See chips.decode)

       Returns:
           dict: {(chip_x, chip_y, x, y): {'k1': [], 'k2': [], 'k3': [], ...}}
    """
    
    return thread_last(chipseq,
                       partial(chips.trim, dates=dateseq),
                       chips.deduplicate,
                       chips.rsort,
                       partial(map, decode_fn),
                       partial(chips.to_numpy, spec_index=spec_index),
                       excepts(ValueError, from_chips, lambda _: []),
                       excepts(AttributeError, partial(locate, locations=locations), lambda _: {}),
                       partial(identify, x=x, y=y))

